#!/bin/bash

########################################################################################
#                                                                                      #
#  contig_info: a BASH script to estimate standard statistics from FASTA contig files  #
#                                                                                      #
#  Copyright (C) 2015,2018,2019  Alexis Criscuolo                                      #
#                                                                                      #
#  This program is free software:  you can redistribute it and/or modify it under the  #
#  terms  of  the GNU  General  Public  License as  published by  the  Free  Software  #
#  Foundation, either version 3 of the License, or (at your option) any later version  #
#                                                                                      #
#  This program is distributed  in the hope that  it will be useful,  but WITHOUT ANY  #
#  WARRANTY;  without even the  implied warranty of  MERCHANTABILITY or FITNESS FOR A  #
#  PARTICULAR PURPOSE. See the GNU General Public License for more details.            #
#                                                                                      #
#  You should have received a copy of the  GNU General Public License along with this  #
#  program. If not, see <http://www.gnu.org/licenses/>.                                #
#                                                                                      #
#  Contact:                                                                            #
#  Institut Pasteur                                                                    #
#  Bioinformatics and Biostatistics Hub                                                #
#  C3BI, USR 3756 IP CNRS                                                              #
#  Paris, FRANCE                                                                       #
#                                                                                      #
#  alexis.criscuolo@pasteur.fr                                                         #
#                                                                                      #
########################################################################################

########################################################################################
#                                                                                      #
# ============                                                                         #
# = VERSIONS =                                                                         #
# ============                                                                         #
#                                                                                      #
  VERSION=1.0.190426ac                                                                 #
# + options -l and -d (i.e. printing sequence lengths and length distribution, resp.)  #
#   are no longer supported                                                            #
# + residue count always computed (option -r discarded)                                #
# + ultrafast residue count (based on tr + wc)                                         #
# + estimating %AT, %GC, L50, L75, L90                                                 #
# + faster estimation of the sequence length statistics (100% awk)                     #
# + ability to read multiple input files                                               #
#                                                                                      #
# VERSION=0.3.180515ac                                                                 #
#                                                                                      #
########################################################################################
  
########################################################################################
#                                                                                      #
#  ================                                                                    #
#  = INSTALLATION =                                                                    #
#  ================                                                                    #
#                                                                                      #
#  Just give the execute permission to the script contig_info.sh with the following    #
#  command line:                                                                       #
#                                                                                      #
#   chmod +x contig_info.sh                                                            #
#                                                                                      #
########################################################################################

########################################################################################
#                                                                                      #
#  ================                                                                    #
#  = MANUAL       =                                                                    #
#  ================                                                                    #
#                                                                                      #
#                                                                                      #
if [ "$1" = "-?" ] || [ $# -lt 1 ]
then
  cat <<EOF

 contig_info v.$VERSION

 USAGE:  contig_info.sh  [options]  <contig_files> 

  where 'options' are:

   -m <int>    minimum contig length;  every contig sequence of length
               shorter than this cutoff will be discarded (default: 1)
   -g <int>    expected  genome  size  for  computing {N,L}G{50,75,90}
               values instead of {N,L}{50,75,90} ones, respectively
   -t          tab-delimited output

EOF
  exit
fi
#                                                                                      #
########################################################################################

########################################################################################
#                                                                                      #
#  ================                                                                    #
#  = FUNCTIONS    =                                                                    #
#  ================                                                                    #
#                                                                                      #
# = randomfile() ====================================================================  #
#   returns a random file name within /tmp/                                            #
#                                                                                      #
randomfile() {
  rdmf=/tmp/$RANDOM; while [ -e $rdmf ]; do rdmf=/tmp/$RANDOM ; done
  echo $rdmf ;
}
#                                                                                      #
########################################################################################

########################################################################################
####                                                                                ####
#### INITIALIZING PARAMETERS AND READING OPTIONS                                    ####
####                                                                                ####
########################################################################################
MIN_CONTIG_LGT=1;
GENOME_SIZE=0;
TSVOUT=false;
while getopts m:g:t option
do
  case $option in
    m)  MIN_CONTIG_LGT=$OPTARG                            ;;
    g)  GENOME_SIZE=$OPTARG                               ;;
    t)  TSVOUT=true                                       ;;
    :)  echo "option $OPTARG : missing argument" ; exit 1 ;;
    \?) echo "$OPTARG : option invalide" ;         exit 1 ;;
  esac
done
if [ $MIN_CONTIG_LGT -lt 1 ]; then echo "   the min contig length threshold must be a positive integer (option -m)" ; exit 1 ; fi
if [ $GENOME_SIZE -lt 0 ];    then echo "   the expected genome size must be a positive integer (option -g)" ;        exit 1 ; fi

########################################################################################
####                                                                                ####
#### CONTIG INFO                                                                    ####
####                                                                                ####
########################################################################################
if $TSVOUT
then
  CSVCAPT="#File\tNseq\tNres\tA\tC\tG\tT\tN\t%A\t%C\t%G\t%T\t%N\t%AT\t%GC\tMin\tQ25\tMed\tQ75\tMax\tAvg\tN50\tN75\tN90\tL50\tL75\tL90";
  [ $GENOME_SIZE -ne 0 ]&&CSVCAPT="$CSVCAPT\tExpSize";
  echo -e "$CSVCAPT" ;
fi
SEQS=$(randomfile); ## txt file with one sequence per line
for INFILE in "$@"
do
  if [ ! -e $INFILE ] || [ ! -f $INFILE ] || [ ! -r $INFILE ]; then continue; fi
    
  tr -d '\15\32' < $INFILE | awk -v thr=$MIN_CONTIG_LGT '/^>/{if(length(s)>=thr)print s;s="";next}{s=s$0}END{if(length(s)>=thr)print s}' | tr '[:lower:]' '[:upper:]' > $SEQS ;
  R=$(tr -d [:cntrl:] < $SEQS | wc -c); S=$(wc -l < $SEQS); AVG=$(bc -l<<<"scale=2;$R/$S" | sed 's/^\./0./');
  A=$(tr -cd A < $SEQS | wc -c); fA=$(bc -l <<<"scale=2;100*$A/$R" | sed 's/^\./0./');
  C=$(tr -cd C < $SEQS | wc -c); fC=$(bc -l <<<"scale=2;100*$C/$R" | sed 's/^\./0./');
  G=$(tr -cd G < $SEQS | wc -c); fG=$(bc -l <<<"scale=2;100*$G/$R" | sed 's/^\./0./');
  T=$(tr -cd T < $SEQS | wc -c); fT=$(bc -l <<<"scale=2;100*$T/$R" | sed 's/^\./0./');
  N=$(tr -cd N < $SEQS | wc -c); fN=$(bc -l <<<"scale=2;100*$N/$R" | sed 's/^\./0./');
  fGC=$(bc -l <<<"scale=2;100*($C+$G)/($A+$C+$G+$T)" | sed 's/^\./0./'); fAT=$(bc -l <<<"scale=2;100-$fGC" | sed 's/^\./0./');
  ER=$R; [ $GENOME_SIZE != 0 ] && ER=$GENOME_SIZE;
  STATS=$(awk '{print length}' $SEQS | sort -rn | awk -v g=$ER '{l[++n]=$0}END{g50=g/2;g75=3*g/4;g90=9*g/10;i=s=n50=n75=n90=0;while(++i<=n&&n90==0){s+=l[i];n50==0&&s>=g50&&n50=l[i]+(l50=i);n75==0&&s>=g75&&n75=l[i]+(l75=i);n90==0&&s>=g90&&n90=l[i]+(l90=i)}print (n50-l50)"\t"(n75-l75)"\t"(n90-l90)"\t"l50"\t"l75"\t"l90"\t"l[1]"\t"l[int(n/4+1)]"\t"l[int(n/2+1)]"\t"l[int(3*n/4+1)]"\t"l[n]}');
  N50=$(cut -f1 <<<"$STATS"); N75=$(cut -f2 <<<"$STATS"); N90=$(cut -f3 <<<"$STATS"); 
  L50=$(cut -f4 <<<"$STATS"); L75=$(cut -f5 <<<"$STATS"); L90=$(cut -f6 <<<"$STATS"); 
  MAX=$(cut -f7 <<<"$STATS"); 
  Q75=$(cut -f8 <<<"$STATS"); Q50=$(cut -f9 <<<"$STATS"); Q25=$(cut -f10 <<<"$STATS"); 
  MIN=$(cut -f11 <<<"$STATS");

  if ! $TSVOUT
  then
    echo ;
    echo "File                           $(basename $INFILE)" ;
    echo ;
    echo "Number of sequences            $S" ;
    echo ;
    echo "Residue counts:" ;
    echo "  Number of A's                $A  $fA %" ;
    echo "  Number of C's                $C  $fC %" ;
    echo "  Number of G's                $G  $fG %" ;
    echo "  Number of T's                $T  $fT %" ;
    echo "  Number of N's                $N  $fN %" ;
    echo "  Total                        $R" ;
    echo ;
    echo "  %AT                          $fAT %" ;
    echo "  %GC                          $fGC %" ;
    echo ;
    echo "Sequence lengths:" ;
    echo "  Minimum                      $MIN" ;
    echo "  Quartile 25%                 $Q25" ;
    echo "  Median                       $Q50" ;
    echo "  Quartile 75%                 $Q75" ;
    echo "  Maximum                      $MAX" ;
    echo "  Average                      $AVG" ;
    echo ;
    echo "Contiguity statistics:" ;
    echo "  N50                          $N50" ;
    echo "  N75                          $N75" ;
    echo "  N90                          $N90" ;
    echo "  L50                          $L50" ;
    echo "  L75                          $L75" ;
    echo "  L90                          $L90" ;
    if [ $GENOME_SIZE -ne 0 ]; then echo "  Expected genome size         $GENOME_SIZE"; fi
    echo ;
  else
    CSVLINE="$(basename $INFILE)\t$S\t$R\t$A\t$C\t$G\t$T\t$N\t$fA%\t$fC%\t$fG%\t$fT%\t$fN%\t$fAT%\t$fGC%\t$MIN\t$Q25\t$Q50\t$Q75\t$MAX\t$AVG\t$N50\t$N75\t$N90\t$L50\t$L75\t$L90";
    [ $GENOME_SIZE -ne 0 ]&&CSVLINE="$CSVLINE\t$ER";
    echo -e "$CSVLINE" ;
  fi
done

rm -f $SEQS ;

exit







 
